@@ -1,6 +1,6 @@
 (* hey emacs, this is OCaml code: -*- tuareg -*- *)
 (* nbd client library in userspace: Python bindings
- * Copyright (C) 2013-2021 Red Hat Inc.
+ * Copyright (C) 2013-2022 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -34,16 +34,15 @@ let generate_python_methods_h () =
   pr "#include <assert.h>\n";
   pr "\n";
   pr "\
-struct py_aio_buffer {
-  Py_ssize_t len;
-  void *data;
-};
-
 extern char **nbd_internal_py_get_string_list (PyObject *);
 extern void nbd_internal_py_free_string_list (char **);
 extern int nbd_internal_py_get_sockaddr (PyObject *,
     struct sockaddr_storage *, socklen_t *);
-extern struct py_aio_buffer *nbd_internal_py_get_aio_buffer (PyObject *);
+extern PyObject *nbd_internal_py_get_aio_view (PyObject *, int);
+extern int nbd_internal_py_init_aio_buffer (PyObject *);
+extern PyObject *nbd_internal_py_get_nbd_buffer_type (void);
+extern PyObject *nbd_internal_py_wrap_errptr (int);
+extern PyObject *nbd_internal_py_get_subview (PyObject *, const char *, size_t);
 
 static inline struct nbd_handle *
 get_handle (PyObject *obj)
@@ -76,9 +75,6 @@ raise_exception ()
   ) ([ "create"; "close";
        "display_version";
        "alloc_aio_buffer";
-       "aio_buffer_from_bytearray";
-       "aio_buffer_to_bytearray";
-       "aio_buffer_size";
        "aio_buffer_is_zero" ] @ List.map fst handle_calls);
 
   pr "\n";
@@ -108,9 +104,6 @@ let generate_python_libnbdmod_c () =
   ) ([ "create"; "close";
        "display_version";
        "alloc_aio_buffer";
-       "aio_buffer_from_bytearray";
-       "aio_buffer_to_bytearray";
-       "aio_buffer_size";
        "aio_buffer_is_zero" ] @ List.map fst handle_calls);
   pr "  { NULL, NULL, 0, NULL }\n";
   pr "};\n";
@@ -143,9 +136,11 @@ PyInit_libnbdmod (void)
     return NULL;
 
   nbd_internal_py_Error = PyErr_NewException (\"nbd.Error\", NULL, NULL);
-  if (nbd_internal_py_Error == NULL)
+  if (PyModule_AddObject (mod, \"Error\", nbd_internal_py_Error) < 0) {
+    Py_XDECREF (nbd_internal_py_Error);
+    Py_DECREF (mod);
     return NULL;
-  PyModule_AddObject (mod, \"Error\", nbd_internal_py_Error);
+  }
 
   return mod;
 }
@@ -165,12 +160,12 @@ let print_python_closure_wrapper { cbname; cbargs } =
   pr "  const struct user_data *data = user_data;\n";
   pr "  int ret = -1;\n";
   pr "\n";
-  pr "  PyGILState_STATE py_save = PyGILState_UNLOCKED;\n";
+  pr "  PyGILState_STATE py_save = PyGILState_Ensure ();\n";
   pr "  PyObject *py_args, *py_ret;\n";
   List.iter (
     function
-    | CBArrayAndLen (UInt32 n, _) ->
-       pr "  PyObject *py_%s = NULL;\n" n
+    | CBArrayAndLen (UInt32 n, _)
+    | CBBytesIn (n, _)
     | CBMutable (Int n) ->
        pr "  PyObject *py_%s = NULL;\n" n
     | _ -> ()
@@ -187,17 +182,13 @@ let print_python_closure_wrapper { cbname; cbargs } =
        pr "    if (!py_e_%s) { PyErr_PrintEx (0); goto out; }\n" n;
        pr "    PyList_SET_ITEM (py_%s, i_%s, py_e_%s);\n" n n n;
        pr "  }\n"
-    | CBBytesIn _
+    | CBBytesIn (n, len) ->
+       pr "  py_%s = nbd_internal_py_get_subview (data->view, %s, %s);\n" n n len;
+       pr "  if (!py_%s) { PyErr_PrintEx (0); goto out; }\n" n
     | CBInt _
     | CBInt64 _ -> ()
     | CBMutable (Int n) ->
-       pr "  PyObject *py_%s_modname = PyUnicode_FromString (\"ctypes\");\n" n;
-       pr "  if (!py_%s_modname) { PyErr_PrintEx (0); goto out; }\n" n;
-       pr "  PyObject *py_%s_mod = PyImport_Import (py_%s_modname);\n" n n;
-       pr "  Py_DECREF (py_%s_modname);\n" n;
-       pr "  if (!py_%s_mod) { PyErr_PrintEx (0); goto out; }\n" n;
-       pr "  py_%s = PyObject_CallMethod (py_%s_mod, \"c_int\", \"i\", *%s);\n" n n n;
-       pr "  Py_DECREF (py_%s_mod);\n" n;
+       pr "  py_%s = nbd_internal_py_wrap_errptr (*%s);\n" n n;
        pr "  if (!py_%s) { PyErr_PrintEx (0); goto out; }\n" n;
     | CBString _
     | CBUInt _
@@ -206,36 +197,35 @@ let print_python_closure_wrapper { cbname; cbargs } =
   ) cbargs;
   pr "\n";
 
-  pr "  py_args = Py_BuildValue (\"(\"";
-  List.iter (
-    function
-    | CBArrayAndLen (UInt32 n, len) -> pr " \"O\""
-    | CBBytesIn (n, len) -> pr " \"y#\""
-    | CBInt n -> pr " \"i\""
-    | CBInt64 n -> pr " \"L\""
-    | CBMutable (Int n) -> pr " \"O\""
-    | CBString n -> pr " \"s\""
-    | CBUInt n -> pr " \"I\""
-    | CBUInt64 n -> pr " \"K\""
-    | CBArrayAndLen _ | CBMutable _ -> assert false
-  ) cbargs;
-  pr " \")\"";
-  List.iter (
-    function
-    | CBArrayAndLen (UInt32 n, _) -> pr ", py_%s" n
-    | CBBytesIn (n, len) -> pr ", %s, (int) %s" n len
-    | CBMutable (Int n) -> pr ", py_%s" n
-    | CBInt n | CBInt64 n
-    | CBString n
-    | CBUInt n | CBUInt64 n -> pr ", %s" n
-    | CBArrayAndLen _ | CBMutable _ -> assert false
-  ) cbargs;
+  let params =
+    List.map (
+      function
+      | CBArrayAndLen (UInt32 n, _) -> "O", sprintf "py_%s" n
+      | CBBytesIn (n, _) -> "O", sprintf "py_%s" n
+      | CBInt n -> "i", n
+      | CBInt64 n -> "L", n
+      | CBMutable (Int n) -> "O", sprintf "py_%s" n
+      | CBString n -> "s", n
+      | CBUInt n -> "I", n
+      | CBUInt64 n -> "K", n
+      | CBArrayAndLen _ | CBMutable _ -> assert false
+    ) cbargs in
+  pr "  py_args = Py_BuildValue (";
+  pr_wrap ',' (fun () ->
+      pr "\"(";
+      List.iter (
+        function
+        | n, _ -> pr "%s" n
+        ) params;
+      pr ")\"";
+      List.iter (
+        function
+        | _, n -> pr ", %s" n
+        ) params);
   pr ");\n";
   pr "  if (!py_args) { PyErr_PrintEx (0); goto out; }\n";
   pr "\n";
-  pr "  py_save = PyGILState_Ensure ();\n";
   pr "  py_ret = PyObject_CallObject (data->fn, py_args);\n";
-  pr "  PyGILState_Release (py_save);\n";
   pr "\n";
   pr "  Py_DECREF (py_args);\n";
   pr "\n";
@@ -261,6 +251,8 @@ let print_python_closure_wrapper { cbname; cbargs } =
     function
     | CBArrayAndLen (UInt32 n, _) ->
        pr "  Py_XDECREF (py_%s);\n" n
+    | CBBytesIn (n, _) ->
+       pr "  Py_XDECREF (py_%s);\n" n
     | CBMutable (Int n) ->
        pr "  if (py_%s) {\n" n;
        pr "    PyObject *py_%s_ret = PyObject_GetAttrString (py_%s, \"value\");\n" n n;
@@ -268,12 +260,12 @@ let print_python_closure_wrapper { cbname; cbargs } =
        pr "    Py_DECREF (py_%s_ret);\n" n;
        pr "    Py_DECREF (py_%s);\n" n;
        pr "  }\n"
-    | CBBytesIn _
     | CBInt _ | CBInt64 _
     | CBString _
     | CBUInt _ | CBUInt64 _ -> ()
     | CBArrayAndLen _ | CBMutable _ -> assert false
   ) cbargs;
+  pr "  PyGILState_Release (py_save);\n";
   pr "  return ret;\n";
   pr "}\n";
   pr "\n"
@@ -281,7 +273,9 @@ let print_python_closure_wrapper { cbname; cbargs } =
 (* Generate the Python binding. *)
 let print_python_binding name { args; optargs; ret; may_set_error } =
   pr "PyObject *\n";
-  pr "nbd_internal_py_%s (PyObject *self, PyObject *args)\n" name;
+  pr "nbd_internal_py_%s (" name;
+  pr_wrap ',' (fun () -> pr "PyObject *self, PyObject *args");
+  pr ")\n";
   pr "{\n";
   pr "  PyObject *py_h;\n";
   pr "  struct nbd_handle *h;\n";
@@ -293,19 +287,20 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | BytesIn (n, _) ->
        pr "  Py_buffer %s = { .obj = NULL };\n" n
     | BytesOut (n, count) ->
-       pr "  char *%s = NULL;\n" n;
+       pr "  PyObject *%s = NULL;\n" n;
        pr "  Py_ssize_t %s;\n" count
     | BytesPersistIn (n, _)
     | BytesPersistOut (n, _) ->
-       pr "  PyObject *%s; /* PyCapsule pointing to struct py_aio_buffer */\n"
-          n;
-       pr "  struct py_aio_buffer *%s_buf;\n" n
+       pr "  PyObject *%s; /* Buffer-like object or nbd.Buffer */\n" n;
+       pr "  PyObject *%s_view = NULL; /* PyMemoryView of %s */\n" n n;
+       pr "  Py_buffer *py_%s; /* buffer of %s_view */\n" n n
     | Closure { cbname } ->
        pr "  struct user_data *%s_user_data = NULL;\n" cbname;
        pr "  PyObject *py_%s_fn;\n" cbname;
-       pr "  nbd_%s_callback %s = { .callback = %s_wrapper,\n"
-         cbname cbname cbname;
-       pr "                         .free = free_user_data };\n"
+       pr "  nbd_%s_callback %s = { " cbname cbname;
+       pr_wrap ',' (fun () ->
+           pr ".callback = %s_wrapper, .free = free_user_data" cbname);
+       pr " };\n"
     | Enum (n, _) -> pr "  int %s;\n" n
     | Flags (n, _) ->
        pr "  uint32_t %s_u32;\n" n;
@@ -340,9 +335,10 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | OClosure { cbname } ->
        pr "  struct user_data *%s_user_data = NULL;\n" cbname;
        pr "  PyObject *py_%s_fn;\n" cbname;
-       pr "  nbd_%s_callback %s = { .callback = %s_wrapper,\n"
-         cbname cbname cbname;
-       pr "                         .free = free_user_data };\n"
+       pr "  nbd_%s_callback %s = { " cbname cbname;
+       pr_wrap ',' (fun () ->
+           pr ".callback = %s_wrapper, .free = free_user_data" cbname);
+       pr " };\n"
     | OFlags (n, _, _) ->
        pr "  uint32_t %s_u32;\n" n;
        pr "  unsigned int %s; /* really uint32_t */\n" n
@@ -350,62 +346,55 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
   pr "\n";
 
   (* Parse the Python parameters. *)
-  pr "  if (!PyArg_ParseTuple (args, \"O\"";
-  List.iter (
-    function
-    | Bool n -> pr " \"p\""
-    | BytesIn (n, _) -> pr " \"y*\""
-    | BytesPersistIn (n, _) -> pr " \"O\""
-    | BytesOut (_, count) -> pr " \"n\""
-    | BytesPersistOut (_, count) -> pr " \"O\""
-    | Closure _ -> pr " \"O\""
-    | Enum _ -> pr " \"i\""
-    | Flags _ -> pr " \"I\""
-    | Fd n | Int n -> pr " \"i\""
-    | Int64 n -> pr " \"L\""
-    | Path n -> pr " \"O&\""
-    | SizeT n -> pr " \"n\""
-    | SockAddrAndLen (n, _) -> pr " \"O\""
-    | String n -> pr " \"s\""
-    | StringList n -> pr " \"O\""
-    | UInt n | UIntPtr n -> pr " \"I\""
-    | UInt32 n -> pr " \"I\""
-    | UInt64 n -> pr " \"K\""
-  ) args;
-  List.iter (
-    function
-    | OClosure _ -> pr " \"O\""
-    | OFlags _ -> pr " \"I\""
-  ) optargs;
-  pr "\n";
-  pr "                         \":nbd_%s\",\n" name;
-  pr "                         &py_h";
-  List.iter (
-    function
-    | Bool n -> pr ", &%s" n
-    | BytesIn (n, _) | BytesPersistIn (n, _)
-    | BytesPersistOut (n, _) -> pr ", &%s" n
-    | BytesOut (_, count) -> pr ", &%s" count
-    | Closure { cbname } -> pr ", &py_%s_fn" cbname
-    | Enum (n, _) -> pr ", &%s" n
-    | Flags (n, _) -> pr ", &%s" n
-    | Fd n | Int n | SizeT n | Int64 n -> pr ", &%s" n
-    | Path n -> pr ", PyUnicode_FSConverter, &py_%s" n
-    | SockAddrAndLen (n, _) -> pr ", &%s" n
-    | String n -> pr ", &%s" n
-    | StringList n -> pr ", &py_%s" n
-    | UInt n | UIntPtr n -> pr ", &%s" n
-    | UInt32 n -> pr ", &%s" n
-    | UInt64 n -> pr ", &%s" n
-  ) args;
-  List.iter (
-    function
-    | OClosure { cbname } -> pr ", &py_%s_fn" cbname
-    | OFlags (n, _, _) -> pr ", &%s" n
-  ) optargs;
+  let params =
+    List.map (
+      function
+      | Bool n -> "p", sprintf "&%s" n, n
+      | BytesIn (n, _) -> "y*", sprintf "&%s" n, sprintf "%s.buf, %s.len" n n
+      | BytesOut (n, count) ->
+         "n", sprintf "&%s" count,
+         sprintf "PyByteArray_AS_STRING (%s), %s" n count
+      | BytesPersistIn (n, _) | BytesPersistOut (n, _) ->
+         "O", sprintf "&%s" n, sprintf "py_%s->buf, py_%s->len" n n
+      | Closure { cbname } -> "O", sprintf "&py_%s_fn" cbname, cbname
+      | Enum (n, _) -> "i", sprintf "&%s" n, n
+      | Flags (n, _) -> "I", sprintf "&%s" n, sprintf "%s_u32" n
+      | Fd n | Int n -> "i", sprintf "&%s" n, n
+      | Int64 n -> "L", sprintf "&%s" n, sprintf "%s_i64" n
+      | Path n -> "O&", sprintf "PyUnicode_FSConverter, &py_%s" n, n
+      | SizeT n -> "n", sprintf "&%s" n, sprintf "(size_t)%s" n
+      | SockAddrAndLen (n, _) ->
+         "O", sprintf "&%s" n,
+         sprintf "(struct sockaddr *) &%s_sa, %s_len" n n
+      | String n -> "s", sprintf "&%s" n, n
+      | StringList n -> "O", sprintf "&py_%s" n, n
+      | UInt n | UIntPtr n -> "I", sprintf "&%s" n, n
+      | UInt32 n -> "I", sprintf "&%s" n, sprintf "%s_u32" n
+      | UInt64 n -> "K", sprintf "&%s" n, sprintf "%s_u64" n
+    ) args in
+  let optparams =
+    List.map (
+      function
+      | OClosure { cbname } -> "O", sprintf "&py_%s_fn" cbname, cbname
+      | OFlags (n, _, _) -> "I", sprintf "&%s" n, sprintf "%s_u32" n
+    ) optargs in
+  let params = params @ optparams in
+  pr "  if (!PyArg_ParseTuple (";
+  pr_wrap ',' (fun () ->
+      pr "args, \"O";
+      List.iter (
+        function
+        | n, _, _ -> pr "%s" n
+        ) params;
+      pr ":nbd_%s\", &py_h" name;
+      List.iter (
+        function
+        | _, n, _ -> pr ", %s" n
+        ) params);
   pr "))\n";
   pr "    goto out;\n";
 
+  (* Two passes over parameters. Any 'goto err' must be in first pass. *)
   pr "  h = get_handle (py_h);\n";
   pr "  if (!h) goto out;\n";
   List.iter (
@@ -432,14 +421,18 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | Bool _ -> ()
     | BytesIn _ -> ()
     | BytesOut (n, count) ->
-       pr "  %s = malloc (%s);\n" n count;
-       pr "  if (%s == NULL) { PyErr_NoMemory (); goto out; }\n" n
-    | BytesPersistIn (n, _) | BytesPersistOut (n, _) ->
-       pr "  %s_buf = nbd_internal_py_get_aio_buffer (%s);\n" n n;
-       pr "  if (!%s_buf) goto out;\n" n;
-       pr "  /* Increment refcount since buffer may be saved by libnbd. */\n";
-       pr "  Py_INCREF (%s);\n" n;
-       pr "  completion_user_data->buf = %s;\n" n
+       pr "  %s = PyByteArray_FromStringAndSize (NULL, %s);\n" n count;
+       pr "  if (%s == NULL) goto out;\n" n
+    | BytesPersistIn (n, _) ->
+       pr "  %s_view = nbd_internal_py_get_aio_view (%s, PyBUF_READ);\n" n n;
+       pr "  if (!%s_view) goto out;\n" n;
+       pr "  py_%s = PyMemoryView_GET_BUFFER (%s_view);\n" n n;
+       pr "  completion_user_data->view = %s_view;\n" n
+    | BytesPersistOut (n, _) ->
+       pr "  %s_view = nbd_internal_py_get_aio_view (%s, PyBUF_WRITE);\n" n n;
+       pr "  if (!%s_view) goto out;\n" n;
+       pr "  py_%s = PyMemoryView_GET_BUFFER (%s_view);\n" n n;
+       pr "  completion_user_data->view = %s_view;\n" n
     | Closure { cbname } ->
        pr "  %s.user_data = %s_user_data = alloc_user_data ();\n" cbname cbname;
        pr "  if (%s_user_data == NULL) goto out;\n" cbname;
@@ -450,7 +443,11 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
        pr "  }\n";
        pr "  /* Increment refcount since pointer may be saved by libnbd. */\n";
        pr "  Py_INCREF (py_%s_fn);\n" cbname;
-       pr "  %s_user_data->fn = py_%s_fn;\n" cbname cbname
+       pr "  %s_user_data->fn = py_%s_fn;\n" cbname cbname;
+       if cbname = "chunk" then (
+         pr "  chunk_user_data->view = nbd_internal_py_get_aio_view (buf, PyBUF_WRITE);\n";
+         pr "  if (!chunk_user_data->view) goto out;\n"
+       )
     | Enum _ -> ()
     | Flags (n, _) -> pr "  %s_u32 = %s;\n" n n
     | Fd _ | Int _ -> ()
@@ -473,35 +470,23 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
   ) args;
   pr "\n";
 
-  (* Call the underlying C function. *)
-  pr "  ret = nbd_%s (h" name;
+  (* Second pass, and call the underlying C function. *)
   List.iter (
     function
-    | Bool n -> pr ", %s" n
-    | BytesIn (n, _) -> pr ", %s.buf, %s.len" n n
-    | BytesOut (n, count) -> pr ", %s, %s" n count
-    | BytesPersistIn (n, _)
-    | BytesPersistOut (n, _) -> pr ", %s_buf->data, %s_buf->len" n n
-    | Closure { cbname } -> pr ", %s" cbname
-    | Enum (n, _) -> pr ", %s" n
-    | Flags (n, _) -> pr ", %s_u32" n
-    | Fd n | Int n -> pr ", %s" n
-    | Int64 n -> pr ", %s_i64" n
-    | Path n -> pr ", %s" n
-    | SizeT n -> pr ", (size_t)%s" n
-    | SockAddrAndLen (n, _) -> pr ", (struct sockaddr *) &%s_sa, %s_len" n n
-    | String n -> pr ", %s" n
-    | StringList n -> pr ", %s" n
-    | UInt n | UIntPtr n -> pr ", %s" n
-    | UInt32 n -> pr ", %s_u32" n
-    | UInt64 n -> pr ", %s_u64" n
+    | BytesPersistOut (n, _) ->
+       pr "  if (nbd_internal_py_init_aio_buffer (%s) < 0) goto out;\n" n
+    | _ -> ()
   ) args;
-  List.iter (
-    function
-    | OClosure { cbname } -> pr ", %s" cbname
-    | OFlags (n, _, _) -> pr ", %s_u32" n
-  ) optargs;
+  pr "  Py_BEGIN_ALLOW_THREADS\n";
+  pr "  ret = nbd_%s (" name;
+  pr_wrap ',' (fun () ->
+      pr "h";
+      List.iter (
+        function
+        | _, _, n -> pr ", %s" n
+        ) params);
   pr ");\n";
+  pr "  Py_END_ALLOW_THREADS\n";
   List.iter (
     function
     | Closure { cbname } -> pr "  %s_user_data = NULL;\n" cbname
@@ -524,26 +509,11 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
   let use_ret = ref true in
   List.iter (
     function
-    | BytesOut (n, count) ->
-       pr "  py_ret = PyBytes_FromStringAndSize (%s, %s);\n" n count;
+    | BytesOut (n, _) ->
+       pr "  py_ret = %s;\n" n;
+       pr "  %s = NULL;\n" n;
        use_ret := false
-    | Bool _
-    | BytesIn _
-    | BytesPersistIn _ | BytesPersistOut _
-    | Closure _
-    | Enum _
-    | Flags _
-    | Fd _ | Int _
-    | Int64 _
-    | Path _
-    | SockAddrAndLen _
-    | SizeT _
-    | String _
-    | StringList _
-    | UInt _
-    | UIntPtr _
-    | UInt32 _
-    | UInt64 _ -> ()
+    | _ -> ()
   ) args;
   if !use_ret then (
     match ret with
@@ -572,8 +542,8 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | BytesIn (n, _) ->
        pr "  if (%s.obj)\n" n;
        pr "    PyBuffer_Release (&%s);\n" n
-    | BytesOut (n, _) -> pr "  free (%s);\n" n
-    | BytesPersistIn _ | BytesPersistOut _ -> ()
+    | BytesOut (n, _) -> pr "  Py_XDECREF (%s);\n" n
+    | BytesPersistIn (n, _) | BytesPersistOut (n, _) -> ()
     | Closure { cbname } ->
        pr "  free_user_data (%s_user_data);\n" cbname
     | Enum _ -> ()
@@ -622,7 +592,7 @@ let generate_python_methods_c () =
   pr " */\n";
   pr "struct user_data {\n";
   pr "  PyObject *fn;    /* Optional pointer to Python function. */\n";
-  pr "  PyObject *buf;   /* Optional pointer to persistent buffer. */\n";
+  pr "  PyObject *view;  /* Optional PyMemoryView of persistent buffer. */\n";
   pr "};\n";
   pr "\n";
   pr "static struct user_data *\n";
@@ -642,8 +612,10 @@ let generate_python_methods_c () =
   pr "  struct user_data *data = user_data;\n";
   pr "\n";
   pr "  if (data) {\n";
+  pr "    PyGILState_STATE py_save = PyGILState_Ensure ();\n";
   pr "    Py_XDECREF (data->fn);\n";
-  pr "    Py_XDECREF (data->buf);\n";
+  pr "    Py_XDECREF (data->view);\n";
+  pr "    PyGILState_Release (py_save);\n";
   pr "    free (data);\n";
   pr "  }\n";
   pr "}\n";
@@ -760,29 +732,70 @@ class Buffer(object):
     '''Asynchronous I/O persistent buffer'''
 
     def __init__(self, len):
-        '''allocate an uninitialized AIO buffer used for nbd.aio_pread'''
+        '''Allocate an uninitialized AIO buffer used for nbd.aio_pread.'''
         self._o = libnbdmod.alloc_aio_buffer(len)
 
     @classmethod
-    def from_bytearray(cls, ba):
-        '''create an AIO buffer from a bytearray'''
-        o = libnbdmod.aio_buffer_from_bytearray(ba)
+    def from_buffer(cls, buf):
+        '''Create an AIO buffer that shares an existing buffer-like object.
+
+        Because the buffer is shared, changes to the original are visible
+        to nbd.aio_pwrite, and changes in nbd.aio_pread are visible to the
+        original.
+        '''
         self = cls(0)
-        self._o = o
+        # Ensure that buf is already buffer-like
+        with memoryview(buf):
+            self._o = buf
+        self._init = True
         return self
 
+    @classmethod
+    def from_bytearray(cls, ba):
+        '''Create an AIO buffer from a bytearray or other buffer-like object.
+
+        If ba is not a buffer, it is tried as the parameter to the
+        bytearray constructor.  Otherwise, ba is copied.  Either way, the
+        resulting AIO buffer is independent from the original.
+        '''
+        return cls.from_buffer(bytearray(ba))
+
+    def to_buffer(self):
+        '''Return a shared view of the AIO buffer contents.
+
+        This exposes the underlying buffer; changes to the buffer are
+        visible to nbd.aio_pwrite, and changes from nbd.aio_pread are
+        visible in the buffer.
+        '''
+        if not hasattr(self, '_init'):
+            self._o = bytearray(len(self._o))
+            self._init = True
+        return self._o
+
     def to_bytearray(self):
-        '''copy an AIO buffer into a bytearray'''
-        return libnbdmod.aio_buffer_to_bytearray(self._o)
+        '''Copy an AIO buffer into a bytearray.
+
+        This copies the contents of an AIO buffer to a new bytearray, which
+        remains independent from the original.
+        '''
+        if not hasattr(self, '_init'):
+            return bytearray(len(self._o))
+        return bytearray(self._o)
 
     def size(self):
-        '''return the size of an AIO buffer'''
-        return libnbdmod.aio_buffer_size(self._o)
+        '''Return the size of an AIO buffer.'''
+        return len(self)
+
+    def __len__(self):
+        '''Return the size of an AIO buffer.'''
+        return len(self._o)
 
     def is_zero(self, offset=0, size=-1):
-        '''
-        Returns true if and only if all bytes in the buffer are zeroes.
-        Note that a freshly allocated buffer is uninitialized, not zero.
+        '''Returns true if and only if all bytes in the buffer are zeroes.
+
+        Note that although a freshly allocated buffer is uninitialized,
+        this will report it as all zeroes, as it will be force-initialized
+        to zero before any code that can access the buffer's contents.
 
         By default this tests the whole buffer, but you can restrict
         the test to a sub-range of the buffer using the optional
@@ -791,18 +804,19 @@ class Buffer(object):
         always returns true.  If size > 0, we check the interval
         [offset..offset+size-1].
         '''
-        return libnbdmod.aio_buffer_is_zero(self._o, offset, size)
+        return libnbdmod.aio_buffer_is_zero(self._o, offset, size,
+                                            hasattr(self, '_init'))
 
 
 class NBD(object):
     '''NBD handle'''
 
     def __init__(self):
-        '''create a new NBD handle'''
+        '''Create a new NBD handle.'''
         self._o = libnbdmod.create()
 
     def __del__(self):
-        '''close the NBD handle and underlying connection'''
+        '''Close the NBD handle and underlying connection.'''
         if hasattr(self, '_o'):
             libnbdmod.close(self._o)
 
@@ -813,31 +827,31 @@ class NBD(object):
       let args =
         List.map (
           function
-          | Bool n -> n, None, None
-          | BytesIn (n, _) -> n, None, None
-          | BytesOut (_, count) -> count, None, None
-          | BytesPersistIn (n, _) -> n, None, Some (sprintf "%s._o" n)
-          | BytesPersistOut (n, _) -> n, None, Some (sprintf "%s._o" n)
-          | Closure { cbname } -> cbname, None, None
-          | Enum (n, _) -> n, None, None
-          | Flags (n, _) -> n, None, None
-          | Fd n | Int n -> n, None, None
-          | Int64 n -> n, None, None
-          | Path n -> n, None, None
-          | SizeT n -> n, None, None
-          | SockAddrAndLen (n, _) -> n, None, None
-          | String n -> n, None, None
-          | StringList n -> n, None, None
-          | UInt n -> n, None, None
-          | UInt32 n -> n, None, None
-          | UInt64 n -> n, None, None
-          | UIntPtr n -> n, None, None
+          | Bool n -> n, None
+          | BytesIn (n, _) -> n, None
+          | BytesOut (_, count) -> count, None
+          | BytesPersistIn (n, _) -> n, None
+          | BytesPersistOut (n, _) -> n, None
+          | Closure { cbname } -> cbname, None
+          | Enum (n, _) -> n, None
+          | Flags (n, _) -> n, None
+          | Fd n | Int n -> n, None
+          | Int64 n -> n, None
+          | Path n -> n, None
+          | SizeT n -> n, None
+          | SockAddrAndLen (n, _) -> n, None
+          | String n -> n, None
+          | StringList n -> n, None
+          | UInt n -> n, None
+          | UInt32 n -> n, None
+          | UInt64 n -> n, None
+          | UIntPtr n -> n, None
         ) args in
       let optargs =
         List.map (
           function
-          | OClosure { cbname } -> cbname, Some "None", None
-          | OFlags (n, _, _) -> n, Some "0", None
+          | OClosure { cbname } -> cbname, Some "None"
+          | OFlags (n, _, _) -> n, Some "0"
         ) optargs in
       let args = args @ optargs in
       pr "    def %s(" name;
@@ -845,21 +859,20 @@ class NBD(object):
           pr "self";
           List.iter (
             function
-            | n, None, _ -> pr ", %s" n
-            | n, Some default, _ -> pr ", %s=%s" n default
+            | n, None -> pr ", %s" n
+            | n, Some default -> pr ", %s=%s" n default
           ) args);
       pr "):\n";
       let longdesc = Str.global_replace py_fn_rex "C<nbd.\\1>" longdesc in
       let longdesc = Str.global_replace py_const_rex "C<" longdesc in
       let longdesc = pod2text longdesc in
-      pr "        '''▶ %s\n\n%s'''\n" shortdesc (String.concat "\n" longdesc);
+      pr "        u'''▶ %s\n\n%s'''\n" shortdesc (String.concat "\n" longdesc);
       pr "        return libnbdmod.%s(" name;
       pr_wrap ',' (fun () ->
           pr "self._o";
           List.iter (
             function
-            | _, _, Some getter -> pr ", %s" getter
-            | n, _, None -> pr ", %s" n
+            | n, _ -> pr ", %s" n
           ) args);
       pr ")\n";
       pr "\n"
