@@ -17,6 +17,8 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  *)
 
+(* For explanation see generator/README.state-machine.md *)
+
 open Printf
 
 type external_event =
@@ -48,9 +50,6 @@ let string_of_external_event = function
   | CmdConnectSocket -> "CmdConnectSocket"
   | CmdIssue -> "CmdIssue"
 
-type location = string * int
-let noloc = ("", 0)
-
 type state = {
   name : string;
   comment : string;
@@ -61,7 +60,7 @@ and parsed_state = {
   prefix : string list;
   display_name : string;
   state_enum : string;
-  loc : location;
+  loc : Utils.location;
   code : string;
   internal_transitions : state list;
   events : (external_event * state) list;
@@ -69,7 +68,8 @@ and parsed_state = {
 
 let default_state = { name = ""; comment = ""; external_events = [];
                       parsed = { prefix = []; display_name = "";
-                                 state_enum = ""; loc = noloc; code = "";
+                                 state_enum = ""; loc = Utils.noloc;
+                                 code = "";
                                  internal_transitions = []; events = [] } }
 
 type state_machine = state_group list
@@ -750,7 +750,7 @@ and reply_state_machine = [
     default_state with
     name = "START";
     comment = "Prepare to receive a reply from the remote server";
-    external_events = [];
+    external_events = [ NotifyRead, "" ];
   };
 
   State {
@@ -878,7 +878,3 @@ and structured_reply_state_machine = [
     external_events = [];
   };
 ]
-
-let string_of_location (file, lineno) = sprintf "%s:%d" file lineno
-let line_directive_of_location (file, lineno) =
-  sprintf "#line %d \"%s\"" lineno file
