@@ -1,4 +1,4 @@
-#!/bin/sh -
+#!/bin/bash -
 # nbd client library in userspace
 # Copyright (C) 2013-2020 Red Hat Inc.
 #
@@ -16,11 +16,21 @@
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 
-set -e
+. ../tests/functions.sh
 
-cd src/$pkg
+set -e
+set -x
+
+requires nbdkit --version
 
 # The -count=1 parameter is the "idiomatic way to bypass test caching".
 # https://golang.org/doc/go1.10#test
 # The -v option enables verbose output.
 $GOLANG test -count=1 -v
+
+# Run the only benchmarks with 10 milliseconds timeout to make sure they
+# do not break by mistake, without overloading the CI. For performance
+# testing run "go test" directly.
+# The -run=- parameter is the way to match no test, assuming that no test or
+# sub test contains "-" in the name.
+$GOLANG test -run=- -bench=. -benchtime=10ms
