@@ -1,5 +1,5 @@
 /* nbd client library in userspace: internal definitions
- * Copyright (C) 2013-2020 Red Hat Inc.
+ * Copyright (C) 2013-2022 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -39,8 +39,8 @@
 
 #include "byte-swapping.h"
 #include "states.h"
+#include "string-vector.h"
 #include "unlocked.h"
-#include "vector.h"
 
 /* Define unlikely macro, but only for GCC.  These are used to move
  * debug and error handling code out of hot paths, making the hot path
@@ -83,8 +83,6 @@ struct command_cb {
   nbd_completion_callback completion;
 };
 
-DEFINE_VECTOR_TYPE (string_vector, char *)
-
 struct nbd_handle {
   /* Unique name assigned to this handle for debug messages
    * (to avoid having to print actual pointers).
@@ -120,8 +118,12 @@ struct nbd_handle {
   uint8_t opt_current; /* 0 or one of NBD_OPT_* */
   struct command_cb opt_cb;
 
-  /* Full info mode. */
-  bool full_info;
+  /* Tweak what OPT_INFO/GO requests. */
+  bool request_block_size; /* default true, for INFO_BLOCK_SIZE */
+  bool full_info; /* default false, for INFO_NAME, INFO_DESCRIPTION */
+
+  /* Sanitization for pread. */
+  bool pread_initialize;
 
   /* Global flags from the server. */
   uint16_t gflags;
@@ -305,6 +307,7 @@ struct nbd_handle {
   struct command *reply_cmd;
 
   bool disconnect_request;      /* True if we've queued NBD_CMD_DISC */
+  bool tls_shut_writes;         /* Used by lib/crypto.c to track disconnect. */
 };
 
 struct meta_context {
@@ -350,7 +353,8 @@ struct command {
   void *data; /* Buffer for read/write */
   struct command_cb cb;
   enum state state; /* State to resume with on next POLLIN */
-  bool data_seen; /* For read, true if at least one data chunk seen */
+  bool initialized; /* For read, true if getting a hole may skip memset */
+  uint32_t data_seen; /* For read, cumulative size of data chunks seen */
   uint32_t error; /* Local errno value */
 };
 
